import json
import csv
import os

from . import utils 

def stringify_all(source):
    return [utils.force_string(v) for v in source]

class FileWriter():
    DEFAULT_OUTPUT = '-'
    DELIMITER = '\n'

    @classmethod
    def args(cls, parser):
        parser.add_argument(
            '--output',
            default=cls.DEFAULT_OUTPUT,
            help='Set target of output (Default stdout)',
        )

    def __init__(self, output=None):
        self.target_name = output
        self.target = None
        self.target_template = None
        self.first_written = False

        self.force_closing_newline = os.environ.get('STREAMLINE_CLOSING_NEWLINE')

    def _output_value(self, entry):
        return utils.force_string(entry.value)

    async def stream(self, source):
        output = ''
        if '{' in self.target_name:
            self.target_template = self.target_name
        else:
            self.target = utils.get_file_io(self.target_name, write=True)
        async for entry in source:
            output = self._output_value(entry)
            if self.target_template:

                file_name = self.target_template.format(input=entry.original_value, index=entry.index)
                with open(file_name, 'w') as target_file:
                    target_file.write(output)
            else:
                if self.force_closing_newline:
                    output = output + self.DELIMITER
                elif self.first_written:
                    output = self.DELIMITER + output
                else:
                    self.first_written = True
                self.target.write(output)
            if hasattr(self.target, 'flush'):
                self.target.flush()

        if self.target and hasattr(self.target, 'close'):
            self.target.close()

class CSVWriter():
    DEFAULT_OUTPUT = '-'

    @classmethod
    def args(cls, parser):
        parser.add_argument('--output', help='Set target of output (Default stdout)', default=cls.DEFAULT_OUTPUT)
        parser.add_argument(
            '--input-column',
            action='store_true',
            default=False,
            help='Automatically add a column for input value',
        )

    def __init__(self, output=DEFAULT_OUTPUT, input_column=False):
        self.target_name = output
        self.input_column = input_column

    def _parse_fields(self, entry):
        if not isinstance(entry.value, dict):
            return None
        else:
            return entry.value.keys()
    
    def _get_values(self, entry, fields):
        if fields is None:
            return stringify_all((entry.original_value, entry.value))

        # Return blanks for all invalid rows
        if not isinstance(entry.value, dict):
            return [''] * len(fields)

        values = [entry.value.get(field, '') for field in fields]
        if self.input_column:
            values.insert(0, entry.original_value)
        return stringify_all(values)

    async def stream(self, source):
        self.target = utils.get_file_io(self.target_name, write=True)
        writer = csv.writer(self.target, 'unix', quoting=csv.QUOTE_MINIMAL)

        fields = None
        header_written = False
        async for entry in source:
            # Write csv column names
            if not header_written:
                fields = self._parse_fields(entry)
                if fields is None:
                    writer.writerow(['input', 'value'])
                else:
                    header = list(fields)
                    if self.input_column:
                        header.insert(0, 'input')
                    writer.writerow(header)
                header_written = True

            # Write values
            entry_values = self._get_values(entry, fields)
            writer.writerow(entry_values)

        if hasattr(self.target, 'close'):
            self.target.close()

class JsonWriter():
    DEFAULT_OUTPUT = '-'

    @classmethod
    def args(cls, parser):
        parser.add_argument(
            '--output',
            default=cls.DEFAULT_OUTPUT,
            help='Set target of output (Default stdout)',
        )

    def __init__(self, output=DEFAULT_OUTPUT):
        self.target_name = output

    async def stream(self, source):
        self.target = utils.get_file_io(self.target_name, write=True)
        self.target.write('[\n')
        first = True
        async for entry in source:
            if not first:
                self.target.write(',\n    ')
            else:
                self.target.write('    ')
            self.target.write(json.dumps(entry.value))
            first = False
        self.target.write('\n]')

        if hasattr(self.target, 'close'):
            self.target.close()


CONSUMERS = {
    'file': FileWriter,
    'csv': CSVWriter,
    'json': JsonWriter,
}

def load_consumer(path):
    if path is None:
        return None
    elif '.' in path:
        Consumer = utils.import_obj(path)
    else:
        Consumer = CONSUMERS.get(path)
    if Consumer is None:
        raise ValueError('Invalid consumer: {}'.format(path))
    return Consumer




class EntryFactory():
    def __init__(self, error_value=None):
        self.error_value = error_value
        self.index = 0

    def __call__(self, value):
        entry = Entry(
            value,
            error_value=self.error_value,
            index=self.index
        )
        self.index += 1
        return entry

class Entry():
    def __init__(self, value, index=None, error_value=None):
        self.index = index
        self.history = [[value]]
        self.errors = []
        self.error_value = error_value

    def push(self, value=None):
        if value is None:
            value = self.value
        self.history.append([value])

    def pop(self):
        if len(self.history) == 1:
            raise ValueError('Attempted to pop an entry history that is only 1 level deep!')
        value = self.value
        self.history = self.history[:-1]
        self.value = value

    def reset(self):
        self.history = [[self.history[0][0]]]

    def collapse(self):
        self.history[-1] = [self.history[-1][-1]]

    def get_history(self):
        return self.history[-1]

    @property
    def original_value(self):
        return self.history[-1][0]

    def get_value(self):
        return self.history[-1][-1]

    def set_value(self, new_value):
        self.history[-1].append(new_value)

    def error(self, e):
        self.errors.append(e)
        self.value = self.error_value

    value = property(get_value, set_value)

def entry_wrap(items):
    return [Entry(item) for item in items]

def entry_unwrap(entries):
    return [entry.value for entry in entries]



from . import utils 
from . import entries

import glob
import json
import csv
import os

class FileReader():
    DELIMITER = '\n'
    DEFAULT_SOURCE = '-'

    @classmethod
    def args(cls, parser):
        parser.add_argument(
            '--input',
            default=cls.DEFAULT_SOURCE,
            dest='source_name',
            help='Set source (Default stdin)',
        )
        parser.add_argument(
            '-k', '--keep-trailing-newline',
            help='Dont automatically trim the ending newline character',
            action='store_true',
            default=False,
        )

    def __init__(self, source_name=DEFAULT_SOURCE, keep_trailing_newline=False):
        self.source_name = source_name
        self.keep_trailing_newline = keep_trailing_newline
        self.source = None

    async def stream(self):
        factory = entries.EntryFactory()
        source = utils.get_file_io(self.source_name)

        ending_delim = False
        try:
            for index, line in enumerate(source):
                ending_delim = line.endswith(self.DELIMITER)
                if ending_delim:
                    line = line.rstrip(self.DELIMITER)
                yield factory(line.rstrip(self.DELIMITER))
        except KeyboardInterrupt as e:
            pass

        if ending_delim and self.keep_trailing_newline:
            yield factory('')

        if hasattr(source, 'close'):
            source.close()

class CSVReader():
    DEFAULT_SOURCE = '-'

    @classmethod
    def args(cls, parser):
        parser.add_argument(
            '--input',
            default=cls.DEFAULT_SOURCE,
            dest='source_name',
            help='Set source (Default stdin)',
        )

    def __init__(self, source_name, **kwargs):
        self.source_name = source_name
        self.source = None

    async def stream(self):
        factory = entries.EntryFactory()
        source = utils.get_file_io(self.source_name)
        reader = csv.DictReader(source)

        for row in reader:
            yield factory(row)

        if hasattr(source, 'close'):
            source.close()

class JsonReader():
    DEFAULT_SOURCE = '-'

    @classmethod
    def args(cls, parser):
        parser.add_argument(
            '--input',
            default=cls.DEFAULT_SOURCE,
            dest='source_name',
            help='Set source (Default stdin)',
        )

    def __init__(self, source_name, **kwargs):
        self.source_name = source_name
        self.source = None

    async def stream(self):
        factory = entries.EntryFactory()
        source = utils.get_file_io(self.source_name)
        content = source.read()
        if hasattr(source, 'close'):
            source.close()

        data = json.loads(content)
        if isinstance(data, list):
            for row in data:
                yield factory(row)
        else:
            yield factory(data)

class MultifileReader():
    DEFAULT_PATTERN= '.'

    @classmethod
    def args(cls, parser):
        parser.add_argument(
            '--pattern',
            default=cls.DEFAULT_PATTERN,
            help='Set directory to list files from',
        )
        parser.add_argument(
            '--metadata',
            default=False,
            action='store_true',
            help='Include file metadata',
        )

    def __init__(self, pattern=None, metadata=False):
        self.include_metadata = metadata
        self.pattern = os.path.expanduser(pattern)
        self.pattern = os.path.abspath(self.pattern)
        if os.path.isdir(self.pattern):
            self.pattern += '/*'

    async def stream(self):
        factory = entries.EntryFactory()
        for path in glob.glob(self.pattern):
            if os.path.isfile(path):
                with open(path, 'r') as f:
                    content = f.read()
                if self.include_metadata:
                    yield factory({
                        'path': path,
                        'content': content,
                    })
                else:
                    yield factory(content)


GENERATORS = {
    'file': FileReader,
    'csv': CSVReader,
    'json': JsonReader,
    'files': MultifileReader,
}

def load_generator(path):
    if path is None:
        return None
    elif '.' in path:
        Generator = utils.import_obj(path)
    else:
        Generator = GENERATORS.get(path)
    if Generator is None:
        raise ValueError('Invalid generator: {}'.format(path))
    return Generator

